/* $Id: main.c,v 1.33 1998/03/17 17:11:36 scott Exp svmcguir $ */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "path.h"
#include "option.h"
#include "mesh.h"
#include "util.h"
#include "inner.h"
#include "euro2.h"
#include "outer_control.h"
#include "blackScholes.h"

#define WEIGHT(time,prices,toPath) transProb(&(md->option.pData),prices,md->option.pData.price[time+1][toPath]) /md->distribution[(time)][(toPath)]

#define PHRASE "smuggle munitions KGB NORAD Albanian Serbian Ortega cracking radar"

void weightMeasure(meshData *md);

void option_error(char *argv[]){
  printf("Usage: %s parameters\n",argv[0]);
  printf("\t-n underliers (n from Broadie and Glasserman)\n");
  printf("\t-r risk-free rate (r)\n");
  printf("\t-D dividend rate (\\delta)\n");
  printf("\t-v variance (\\sigmna^2)\n");
  printf("\t-T time to maturity (T)\n");
  printf("\t-K strikeprice (K)\n");
  printf("\t-S initial price (S_0)\n");
  printf("\t-N repetitions (N)\n");
  printf("\t-b paths in mesh (b)\n");
  printf("\t-p paths in path estimate (n_p)\n");
  printf("\t-d time slices (d)\n");
  printf("\t-f number of standard deviations kept in filtered average\n");
  printf("\t-c number of paths in a cluster\n");
  printf("\t-I which inner control (0 = no control)\n");
  printf("\nRates and variance are for T=1.\n");
  printf("All parameters are required.\n\n");

  exit(1);
}

int main(int argc, char *argv[]){
  meshData md; 
  int i,j;
  int time, path, underlier, reps;
  int mac, pac, cmac, cpac; /* variouscounters */
  double  *me, *pe, *cme, *cpe; /* Pointers for mesh estimate arrays */
  double  ma, mv, pa, pv, cma, cmv, cpa, cpv;
  double  nma, npa, nmv, npv, ncma, ncmv, ncpa, ncpv;

  /* various parameters and paramter parsing things */
  const char *options = "n:r:D:v:T:K:S:N:b:p:d:f:c:I:";
  int n,N,b,p,d,option,count=0,c,I;
  double r,D,v,K,S,f,T;

  int toPath, fromPath;
  FILE *weightfile;    
  double weightSum, weightSqSum; 

  /* Variables for pricing Europeans, whatever they are worth */
  double europe_value;
  double *euro_price;

  /* Arrays for Outer controls */
  double *Qhat, **uhat, *u;
  double Qbar;

  /* - - - - - Read command line parameters - - - - - */
  while((option = getopt(argc,argv,options)) != EOF){
    switch(option){
    case('I'):  
      sscanf(optarg,"%i",&I);
      count++;
      break;
    case('n'):  
      sscanf(optarg,"%i",&n);
      count++;
      break;
    case('r'):
      sscanf(optarg,"%lf",&r);
      count++;
      break;
    case('D'):
      sscanf(optarg,"%lf",&D);
      count++;
      break;
    case('v'):
      sscanf(optarg,"%lf",&v);
      count++;
      break;
    case('T'):
      sscanf(optarg,"%lf",&T);
      count++;
      break;
    case('K'):
      sscanf(optarg,"%lf",&K);
      count++;
      break;
    case('S'):
      sscanf(optarg,"%lf",&S);
      count++;
      break;
    case('N'):
      sscanf(optarg,"%i",&N);
      count++;
      break;
    case('b'):
      sscanf(optarg,"%i",&b);
      count++;
      break;
    case('p'):
      sscanf(optarg,"%i",&p);
      count++;
      break;
    case('d'):
      sscanf(optarg,"%i",&d);
      count++;
      break;
    case('c'):
      sscanf(optarg,"%i",&c);
      count++;
      break;
    case('f'):
      sscanf(optarg,"%lf",&f);
      count++;
      break;
    default:
      option_error(argv);
    }
  }

  printf("Number of arguments = %i\n",count);
  if(count != 14){
    option_error(argv);
  }

  printf("Values input \n");
  printf("Number of underliers n = %i\n",n);
  printf("Final time T = %lf\n",T);
  printf("Total number of repetitions N = %i\n",N);
  printf("Number of mesh points b = %i\n",b);
  printf("Number of paths p = %i\n",p);
  printf("Number of time slices d = %i\n",d);
  printf("Risk-free interest rate r = %lf\n",r);
  printf("Dividend rate D = %lf\n",D);
  printf("Variance v = %lf\n",v);
  printf("Strike price K = %lf\n",K);
  printf("Initial strike price S = %lf\n",S);
  printf("Number of paths in a cluster c = %i\n",c);
  printf("Number of standard deviations in filtered average f = %lf\n",f);
  printf("Using control I = %i\n",I);


/* - - - - -  Allocating memory for mesh estimates for each mesh - - - - -*/
  allocate1dDouble(&me,N);   /* mesh estimate */
  allocate1dDouble(&pe,N);   /* path estimate */
  allocate1dDouble(&cme,N);  /* controlled mesh estimate */
  allocate1dDouble(&cpe,N);  /* controlled path estimate */

/* - - - - -Array for pricing Europeans as needed in outer control- - - - - */
  allocate1dDouble(&euro_price,d+1);

  /* - - - - - Allocate memory for Outer controls - - - - - */
  allocate1dDouble(&Qhat, N);
  allocate2dDouble(&uhat, 2,N);
  allocate1dDouble(&u, 2);

  /* --------------------------------------------------------*/
  /* The parameters that can be set before memory allocation */
  /* times, paths and underliers required to come before allocation. */
  /* Later, these should go in a config file. */
  md.option.pData.times = d + 1;
  md.option.pData.paths = p;
  md.option.pData.underliers = n;
  md.option.pData.clusterSize = c;
  md.option.strikePrice = K;
  md.option.riskFreeRate = r*(double)T/(double)d;
  md.meshPaths = b;
  md.option.pData.meshPaths = b;
  md.tolerance = 1.0;
  strcpy(md.option.pData.phrase,PHRASE);

  /* Allocate the memory */
  pathDataAllocate(&(md.option.pData));
  optionDataAllocate(&(md.option));
  meshDataAllocate(&md);

  /* Fill the covariance matrix and mean vector */
  /* In our case, the covariance matrix is diagonal with the variance */
  /* in all diagonal entries. */
  for(i=0;i<md.option.pData.underliers;i++){
    for(j=0;j<md.option.pData.underliers;j++){
      if(i == j){
        md.option.pData.covariance[i][j] = v*(double)T/(double)d;
        md.option.pData.growthRate[i] = r*(double)T/(double)d;
        md.option.pData.dividendRate[i] = D*(double)T/(double)d;
      }else{
        md.option.pData.covariance[i][j] = 0;
      }
    }
  }

  /* Sets up internal Random Number Generator stuff */
  /* Must come after the means and covariances ar defined */
  pathParamsInit(&(md.option.pData));

  /* On each path set all the initial prices */
  for(path=0; path<md.option.pData.meshPaths; path++){
    for(underlier=0; underlier < md.option.pData.underliers; underlier++){
      md.option.pData.price[0][path][underlier] = S;
    }
  }

  weightfile = fopen("weightData","a");
  fprintf(weightfile,"#n = %i\n",n);
  fprintf(weightfile,"#T = %i\n",T);
  fprintf(weightfile,"#N = %i\n",N);
  fprintf(weightfile,"#b = %i\n",b);
  fprintf(weightfile,"#p = %i\n",p);
  fprintf(weightfile,"#d = %i\n",d);
  fprintf(weightfile,"#r = %lf\n",r);
  fprintf(weightfile,"#D = %lf\n",D);
  fprintf(weightfile,"#v = %lf\n",v);
  fprintf(weightfile,"#K = %lf\n",K);
  fprintf(weightfile,"#S = %lf\n",S);
  fprintf(weightfile,"#c = %i\n",c);
  fprintf(weightfile,"#f = %lf\n",f);
  fprintf(weightfile,"$I = %i\n",I);

  /* Initialize means and variances for various estimates */
  ma = 0;
  pa = 0;
  mv = 0;
  pv = 0;
  cma  = 0;
  cpa = 0;
  cmv = 0;
  cpv = 0;
  for(reps=0;reps<N;reps++){ 

    pathDataFill(&(md.option.pData));

/*    pathDataOut(&(md.option.pData),1.0*reps);  */

    optionDataFill(&(md.option));

    meshSetFinalValue(&md);

    /*     meshDataFill(&md);  */

    /* Store value of mesh estimate for current repetition */
     me[reps] = md.valueIfHeld[0][0];  

     /*     pe[reps] = meshPathEstimate(&md); */

     ma += me[reps]; 
     mv += me[reps]*me[reps]; 
     pa += pe[reps]; 
     pv += pe[reps]*pe[reps]; 

    /* Apply inner controls */
    innerDataFill(&md,I);

    /* store values of controlled mesh and path estimates */
    cme[reps] = md.newValueIfHeld[0][0];
    cpe[reps] = innerControlledPathEstimate(&md,I);

    cma += cme[reps];
    cmv += cme[reps]*cme[reps];
    cpa += cpe[reps];
    cpv += cpe[reps]*cpe[reps];

    printf("rep=%d: cme=%lf\t cpe=%lf\n",reps,cme[reps],cpe[reps]);

    weightMeasure(&md); 
    for(time=0;time<d;time++){   
      fprintf(weightfile,"%lg\t%lg\t%lg%lg\n",(double)time,md.weightMeasure1[time],md.weightMeasure2[time],md.weightMeasure3[time]);   
    }   

    /* - - - - - Outer control section- - - - -*/
/*     euro2(&md,0.01,euro_price);  */
/*     for(time=0; time < d+1; time++){  */
/*       printf("euro_price[%i] = %lf\n",time,euro_price[time]);  */
/*     }   */

    
    /* compute mesh estimates of u1 and/or u2 */ 
    /*uhat[1][reps] = OuterCtlEuroMaxMeshValue(&md,  r,  (double)T);*/
  }

  /* compute independent value of u1 and/or u2 */
  /*   u[1] = EuropeanMaxOption(&md, r, (double)(T)); */
  /*   u[2] = EuropeanMaxOption(&md, r, 2.0*(double)(T)/3.0); */

  /* Compute outer control estimate  */
  /*   Qbar = Compute_Outer_Control_Estimate(1, N, Qhat, uhat, u); */

  printf("\n");
  ma /= N;
  mv = mv/N - ma*ma;
  pa /= N;
  pv = pv/N - pa*pa;
  cma /= N;
  cmv = cmv/N - cma*cma;
  cpa /= N;
  cpv = cpv/N - cpa*cpa;
  printf("%lf\t%lf\n",cma,cpa);
  printf("%lf\t%lf\n",sqrt(cmv),sqrt(cpv));
  printf("%lf\t%lf\n",sqrt(cmv/N),sqrt(cpv/N));

  nma = 0;
  npa = 0;
  nmv = 0;
  npv = 0;
  ncma  = 0;
  ncpa = 0;
  ncmv = 0;
  ncpv = 0;  
  mac = 0;
  pac = 0;
  cmac = 0;
  cpac = 0;
  for(reps=0;reps<N;reps++){
    if(fabs(me[reps] - ma) < f*sqrt(mv)){
      nma += me[reps];
      nmv += me[reps]*me[reps];
      mac++;
    }
    if(fabs(pe[reps] - pa) < f*sqrt(pv)){
      npa += pe[reps];
      npv += pe[reps]*pe[reps];
      pac++;
    }
    if(fabs(cme[reps] - cma) < f*sqrt(cmv)){
      ncma += cme[reps];
      ncmv += cme[reps]*cme[reps];
      cmac++;
    }
    if(fabs(cpe[reps] - cpa) < f*sqrt(cpv)){
      ncpa += cpe[reps];
      ncpv += cpe[reps]*cpe[reps];
      cpac++;
    }
    

  }

  printf("\n");
  nma /= mac;
  npa /= pac;
  nmv = nmv/mac - nma*nma;
  npv = npv/pac - npa*npa;
  ncma /= cmac;
  ncpa /= cpac;
  ncmv = ncmv/cmac - ncma*ncma;
  ncpv = ncpv/cpac - ncpa*ncpa;  
  printf("%lf\t%lf\n",ncma,ncpa);
  printf("%lf\t%lf\n",sqrt(ncmv),sqrt(ncpv));
  printf("%lf\t%lf\n",sqrt(ncmv/N),sqrt(ncpv/N));


    

  printf("Value of mesh estimate (Qbar) %lf\n",Qbar);

  fclose(weightfile);

}

void weightMeasure(meshData *md){
  int time, fromPath, toPath;
  double weightSum, weightSqSum, maxWeight, weight;

  for(time=0; time <md->option.pData.times-1; time++){
    md->weightMeasure1[time] = 0;
    md->weightMeasure2[time] = 0;
    md->weightMeasure3[time] = 0;
    for(fromPath=0; fromPath<md->meshPaths; fromPath++){
      weightSum = 0;
      weightSqSum = 0;
      maxWeight = 0;
      for(toPath=0; toPath<md->meshPaths; toPath++){
	weight = WEIGHT(time,md->option.pData.price[time][fromPath],toPath);
	weightSum += weight;
	weightSqSum += weight*weight;
	if(weight > maxWeight) maxWeight = weight;
      }
      md->weightMeasure1[time] += weightSum*weightSum/weightSqSum;
      md->weightMeasure2[time] += weightSum/maxWeight;
      md->weightMeasure3[time] += weightSum;

/*       printf("%d\t%d\t %lg\n",time,fromPath,  */
/* 	     weightSum/maxWeight);  */


    }
    md->weightMeasure1[time] /= md->meshPaths;
    md->weightMeasure2[time] /= md->meshPaths;
    md->weightMeasure3[time] /= md->meshPaths;

  }
}
      
